/*
 * xrick/src/draw.c
 *
 * This source code file is part of the xrick distribution, and as such is
 * copyrighted. Please read the README file in the root directory of the
 * distribution for information about copyrights, trademarks, license,
 * disclaimers, etc.
 */

/*
 * NOTES
 *
 * This is the only file which acceses the video. Anything calling d_*
 * function should be video-independant.
 *
 * draw.c draws into a 320x200 or 0x0140x0xc8 8-bits depth frame buffer,
 * using the CGA 2 bits color codes. It is up to the video to figure out
 * how to display the frame buffer. Whatever draw.c does, does not show
 * until the screen is explicitely refreshed.
 *
 * The "screen" is the whole 0x0140 by 0x00c8 screen, coordinates go from
 * 0x0000,0x0000 to 0x013f,0x00c7.
 *
 * The "map" is a 0x0100 by 0x0140 rectangle that represents the active
 * game area.
 *
 * Relative to the screen, the "map" is located at 0x0020,-0x0040 : the
 * "map" is composed of two hidden 0x0100 by 0x0040 rectangles (one at the
 * top and one at the bottom) and one visible 0x0100 by 0x00c0 rectangle (in
 * the middle).
 *
 * The "map screen" is the visible rectangle ; it is a 0x0100 by 0xc0
 * rectangle located at 0x0020,0x00.
 *
 * Coordinates can be relative to the screen, the map, or the map screen.
 *
 * Coordinates can be expressed in pixels. When relative to the map or the
 * map screen, they can also be expressed in tiles, the map being composed
 * of rows of 0x20 tiles of 0x08 by 0x08 pixels.
 */

#include "system.h"
#include "global.h"
#include "draw.h"

#include "video.h"
#include "videoex.h"
#include "sprites.h"
#include "tiles.h"
#include "maps.h"
#include "rects.h"


/*
 * counters positions (pixels, screen)
 */
#define D_CNT_SCORE_X 0x28
#define D_CNT_BULLETS_X 0x68
#define D_CNT_BOMBS_X 0xA8
#define D_CNT_LIVES_X 0xE8
#define D_CNT_Y 0x08


/*
 * global vars
 */
U8 *d_fbp;         /* frame buffer pointer */
U8 *d_tllst;       /* tiles list pointer */
U16 d_filter;      /* CGA colors filter */
U8 d_tlbank;       /* tile number offset */


/*
 * Set the frame buffer pointer
 *
 * x, y: position (pixels, screen)
 */
void
d_setfbp(U16 x, U16 y)
{
  d_fbp = v_fb + x + y * V_WIDTH;
}


/*
 * Clip to map screen
 *
 * x, y: position (pixels, map) CHANGED clipped
 * width, height: dimension CHANGED clipped
 * return: G_TRUE if fully clipped, G_FALSE if still (at least partly) visible
 */
U8
d_clipms(S16 *x, S16 *y, U16 *width, U16 *height)
{
  if (*x < 0) {
    if (*x + *width < 0)
      return G_TRUE;
    else {
      *width += *x;
      *x = 0;
    }
  }
  else {
    if (*x > 0x0100)
      return G_TRUE;
    else if (*x + *width > 0x0100) {
      *width = 0x0100 - *x;
    }
  }

  if (*y < D_XYMAP_SCRTOP) {
    if ((*y + *height) < D_XYMAP_SCRTOP)
      return G_TRUE;
    else {
      *height += *y - D_XYMAP_SCRTOP;
      *y = D_XYMAP_SCRTOP;
    }
  }
  else {
    if (*y >= D_XYMAP_HBTOP)
      return G_TRUE;
    else if (*y + *height > D_XYMAP_HBTOP)
      *height = D_XYMAP_HBTOP - *y;
  }

  return G_FALSE;
}


/*
 * Draw a list of tiles onto the frame buffer
 * start at position indicated by fbp ; at the end of each (sub)list,
 * perform a "carriage return + line feed" i.e. go back to the initial
 * position then go down one tile row (8 pixels)
 *
 * ASM 1e33
 * d_fbp: CHANGED (see above)
 * d_tllst: CHANGED points to the element following 0xfe/0xff end code
 */
void
d_tl_lst(void)
{
  U8 *t;

  t = d_fbp;
  while (d_tl_sblst() != 0xfe) {  /* draw sub-list */
    t += 8 * V_WIDTH;  /* go down one tile i.e. 8 lines */
    d_fbp = t;
  }
}


/*
 * Draw a sub-list of tiles onto the frame buffer
 * start at position indicated by fbp ; leave fbp pointing to the next
 * tile to the right of the last tile drawn
 *
 * ASM 1e41
 * d_fpb: CHANGED (see above)
 * d_tllst: CHANGED points to the element following 0xfe/0xff end code
 * returns: end code (0xfe : end of list ; 0xff : end of sub-list)
 */
U8
d_tl_sblst()
{
  U8 i;

  i = *(d_tllst++);
  while (i != 0xff && i != 0xfe) {  /* while not end */
    d_tile(i);  /* draw tile */
    i = *(d_tllst++);
  }
  return i;
}


/*
 * Draw a tile
 * at position indicated by fbp ; leave fbp pointing to the next tile
 * to the right of the tile drawn
 *
 * ASM 1e6c
 * tlnbr: tile number
 * d_filter: CGA colors filter
 * d_fbp: CHANGED (see above)
 */
void
d_tile(U8 tlnbr)
{
  U8 i, k, *f;

#ifdef GFXPC
  U16 x;
#endif

#ifdef GFXST
  U32 x;
#endif

  f = d_fbp;  /* frame buffer */
  for (i = 0; i < 8; i++) {  /* for all 8 pixel lines */
    
#ifdef GFXPC
    x = tl_tiles[d_tlbank][tlnbr][i] & d_filter;
    /* 
     * tiles / perform the transformation from CGA 2 bits
     * per pixel to frame buffer 8 bits per pixels
     */
    for (k = 8; k--; x >>= 2)
      f[k] = x & 3;
    f += V_WIDTH;  /* next line */
#endif

#ifdef GFXST
  x = tl_tiles[d_tlbank][tlnbr][i];
  /* 
   * tiles / perform the transformation from ST 4 bits
   * per pixel to frame buffer 8 bits per pixels
   */
  for (k = 8; k--; x >>= 4)
    f[k] = x & 0x0f;
  f += V_WIDTH;  /* next line */
#endif

  }

  d_fbp += 8;  /* next tile */
}

/*
 * Draw a sprite
 *
 * ASM 1a09
 * pln: sprites plane
 * nbr: sprite number
 * x, y: sprite position (pixels, screen)
 * d_fbp: CHANGED
 */
#ifdef GFXPC
void
d_sprite(U8 pln, U8 nbr, U16 x, U16 y)
{
  U8 i, j, k, *f;
  U16 xm, xp;

  d_setfbp(x, y);

  for (i = 0; i < 4; i++) {  /* for each tile column */
    f = d_fbp;  /* frame buffer */
    for (j = 0; j < 0x15; j++) {  /* for each pixel row */
      xm = spr_sprites[pln][nbr][i][j].mask;  /* mask */
      xp = spr_sprites[pln][nbr][i][j].pict;  /* picture */
      /* 
       * sprites / perform the transformation from CGA 2 bits
       * per pixel to frame buffer 8 bits per pixels
       */
      for (k = 8; k--; xm >>= 2, xp >>= 2)
	f[k] = (f[k] & (xm & 3)) | (xp & 3);
      f += V_WIDTH;
    }
    d_fbp += 8;
  }
}
#endif


/*
 * Draw a sprite
 *
 * foobar
 */
#ifdef GFXST
void
d_sprite(U8 nbr, U16 x, U16 y)
{
  U8 i, j, k, *f;
  U16 g;
  U32 d;

  d_setfbp(x, y);
  g = 0;
  for (i = 0; i < 0x15; i++) { /* rows */
    f = d_fbp;
    for (j = 0; j < 4; j++) { /* cols */
      d = spr_sprites[nbr][g++];
      for (k = 8; k--; d >>= 4)
	if (d & 0x0f) f[k] = (f[k] & 0xf0) | (d & 0x0f);
      f += 8;
    }
    d_fbp += V_WIDTH;
  }
}
#endif


/*
 * Draw a sprite
 *
 * NOTE two pixels too far to the right? also horizontal clipping isn't
 * in place but is it needed? must document. no cheat mode. 
 */
#ifdef GFXST
void
d_sprite2(U8 nbr, U16 x, U16 y)
{
  U8 *f, dx;
  S8 r, k, c, i, g;
  U32 d;
  S16 xmap, ymap;
  U16 cmax, rmax;

  xmap = x & 0xfff8;
  ymap = y;
  cmax = 0x20;  /* width, 4 tile columns, 8 pixels each */
  rmax = 0x15;  /* height, 15 pixels */
  dx = x - xmap;
  if (d_clipms(&xmap, &ymap, &cmax, &rmax))  /* don't draw if fully clipped */
    return;

  /* get back to screen */
  d_setfbp(xmap - D_XYMAP_SCRLEFT, ymap - D_XYMAP_SCRTOP);
  xmap >>= 3;
  cmax >>= 3;

  g = (dx == 0 ? 0 : -1);

  for (r = 0; r < rmax; r++) { /* rows */
    f = d_fbp + (dx == 0 ? 0x18 : 0x20);
    i = (dx == 0 ? 7 : dx);
    k = -1;
    for (c = (dx == 0 ? 3 : 4); c >= 0; c--) { /* cols */
      if (!g_cheat3 && /*(c >= cmax) ||*/
	  (map_eflg[map_map[(ymap + r)>>3][xmap + c]] & MAP_EFLG_FGND)) {
	if (k < 0)
	  k = 7 - i;
	d = spr_sprites[nbr][g + c] >> ((7 - k)*4);
      }
      else {
	while (i >= 0) {
	  if (k < 0) {
	    /*if (c == 0) break;*/
	    k = 7;
	    d = spr_sprites[nbr][g + c];
	  }
	  if (d & 0x0f) f[i] = (f[i] & 0xf0) | (d & 0x0f);
	  if (g_cheat3) f[i] |= 0x10;

	  d >>= 4;
	  k--;
	  i--;
	}
      }
      i = 7;
      f -= 8;
    }
    d_fbp += V_WIDTH;
    g += 4;
  }
}
#endif


/*
 * Draw a sprite
 * align to tile column, determine plane automatically, and clip
 *
 * nbr: sprite number
 * x, y: sprite position (pixels, map).
 * d_fbp: CHANGED
 */
#ifdef GFXPC
void
d_sprite2(U8 nbr, U16 x, U16 y)
{
  U8 k, *f, pln, c, r;
  U16 cmax, rmax;
  U16 xm, xp;
  S16 xmap, ymap;

  pln = (x & 0x06) >> 1;  /* calc plane */

  /* align to tile column, prepare map coordinate and clip */
  xmap = x & 0xfff8;
  ymap = y;
  cmax = 0x20;  /* width, 4 tile columns, 8 pixels each */
  rmax = 0x15;  /* height, 15 pixels */
  if (d_clipms(&xmap, &ymap, &cmax, &rmax))  /* don't draw if fully clipped */
    return;

  /* get back to screen */
  d_setfbp(xmap - D_XYMAP_SCRLEFT, ymap - D_XYMAP_SCRTOP);
  xmap >>= 3;
  cmax >>= 3;

  /* draw */
  for (c = 0; c < cmax; c++) {  /* for each tile column */
    f = d_fbp;
    for (r = 0; r < rmax; r++) {  /* for each pixel row */
      /* check that tile is not hidden behind foreground */
      if (g_cheat3 || 
	  !(map_eflg[map_map[(ymap + r) >> 3][xmap + c]] & MAP_EFLG_FGND)) {
	xm = spr_sprites[pln][nbr][c][r].mask;  /* mask */
	xp = spr_sprites[pln][nbr][c][r].pict;  /* picture */
	/* 
	 * sprites / perform the transformation from CGA 2 bits
	 * per pixel to frame buffer 8 bits per pixels
	 */
	for (k = 8; k--; xm >>= 2, xp >>= 2) {
	  f[k] = ((f[k] & (xm & 3)) | (xp & 3));
	  if (g_cheat3) f[k] |= 4;
	}
      }
      f += V_WIDTH;
    }
    d_fbp += 8;
  }
}
#endif


/*
 * Redraw the map behind a sprite
 * align to tile column and row, and clip
 *
 * x, y: sprite position (pixels, map).
 */
void
d_sprite_back(U16 x, U16 y)
{
  U8 r, c;
  U16 rmax, cmax;
  S16 xmap, ymap;
  U16 xs, ys;

  /* aligne to column and row, prepare map coordinate, and clip */
  xmap = x & 0xfff8;
  ymap = y & 0xfff8;
  cmax = (x - xmap == 0 ? 0x20 : 0x28);  /* width, 4 tl cols, 8 pix each */
  rmax = (y & 0x04) ? 0x20 : 0x18;  /* height, 3 or 4 tile rows */
  if (d_clipms(&xmap, &ymap, &cmax, &rmax))  /* don't draw if fully clipped */
    return;

  /* get back to screen */
  xs = xmap - D_XYMAP_SCRLEFT;
  ys = ymap - D_XYMAP_SCRTOP;
  xmap >>= 3;
  ymap >>= 3;
  cmax >>= 3;
  rmax >>= 3;

  /* draw */
  for (r = 0; r < rmax; r++) {  /* for each row */
    d_setfbp(xs, ys + r * 8);
    for (c = 0; c < cmax; c++) {  /* for each column */
      d_tile(map_map[ymap + r][xmap + c]);
    }
  }
}


/*
 * Draw entire map screen background tiles onto frame buffer.
 *
 * ASM 0af5, 0a54
 */
void
d_map(void)
{
  U8 i, j;

  d_tlbank = map_tlbank;

  for (i = 0; i < 0x18; i++) {  /* 0x18 rows */
    d_setfbp(0x20, (i*8));
    for (j = 0; j < 0x20; j++)  /* 0x20 tiles per row */
      d_tile(map_map[i + 8][j]);
  }
}


/*
 * Draw counters
 *
 * ASM 0309
 */
void
d_cnt(void)
{
  S8 i;
  U32 sv;
  static U8 s[7] = {0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0xfe};

  d_tlbank = 0;

  for (i = 5, sv = g_score; i >= 0; i--) {
    s[i] = 0x30 + (U8)(sv % 10);
    sv /= 10;
  }
  d_tllst = s;

  d_setfbp(D_CNT_SCORE_X, D_CNT_Y);
  d_tl_lst();

  d_setfbp(D_CNT_BULLETS_X, D_CNT_Y);
  for (i = 0; i < g_bullets; i++)
    d_tile(TL_ICON_BULLET);

  d_setfbp(D_CNT_BOMBS_X, D_CNT_Y);
  for (i = 0; i < g_bombs; i++)
    d_tile(TL_ICON_BOMB);

  d_setfbp(D_CNT_LIVES_X, D_CNT_Y);
  for (i = 0; i < g_lives; i++)
    d_tile(TL_ICON_RICK);
}


/*
 * Draw status info
 */
void
d_status(void)
{
  d_setfbp(0x00, D_CNT_Y);
  d_tile(g_cheat1 ? 'T' : '@');
  d_setfbp(0x08, D_CNT_Y);
  d_tile(g_cheat2 ? 'N' : '@');
  d_setfbp(0x10, D_CNT_Y);
  d_tile(g_cheat3 ? 'V' : '@');
}


/*
 * Clear counters
 */
void
d_cnt_clr(void)
{
  U8 i;

  d_tlbank = map_tlbank;
  d_setfbp(D_CNT_SCORE_X, D_CNT_Y);
  for (i = 1; i < 0x1f; i++)
    d_tile(map_map[MAP_ROW_SCRTOP + 1][i]);
}


/*
 * Update bullets/bombs counters
 *
 * FIXME get rid of that one
 *
 * ASM 0200
 */
U8
d_cnt_updbb(void)
{
  U8 i, refresh = 0;
  static U8 bombs = 6;
  static U8 bullets = 6;

  d_tlbank = 0;

  if (bullets != g_bullets) {
    bullets = g_bullets;
    d_setfbp(D_CNT_BULLETS_X, D_CNT_Y);
    for (i = 0; i < g_bullets; i++)
      d_tile(TL_ICON_BULLET);
    refresh = 1;
  }

  if (bombs != g_bombs) {
    bombs = g_bombs;
    d_setfbp(D_CNT_BOMBS_X, D_CNT_Y);
    for (i = 0; i < g_bombs; i++)
      d_tile(TL_ICON_BOMB);
    refresh = 1;
  }

  return refresh;
}

/*
 * Draw a picture
 */
#ifdef GFXST
void
d_pic(U16 x, U16 y, U16 w, U16 h, U32 *pic)
{
  U8 *f;
  U16 i, j, k, pp;
  U32 v;

  d_setfbp(x, y);
  pp = 0;

  for (i = 0; i < h; i++) { /* rows */
    f = d_fbp;
    for (j = 0; j < w; j += 8) {  /* cols */
      v = pic[pp++];
      for (k = 8; k--; v >>=4)
	f[k] = v & 0x0f;
      f += 8;
    }
    d_fbp += V_WIDTH;
  }
}
#endif


/* eof */
