/*
 * xrick/src/mainloop.c
 *
 * This source code file is part of the xrick distribution, and as such is
 * copyrighted. Please read the README file in the root directory of the
 * distribution for information about copyrights, trademarks, license,
 * disclaimers, etc.
 */

#include <stdlib.h>

#include "system.h"
#include "global.h"
#include "mainloop.h"

#include "video.h"
#include "draw.h"
#include "maps.h"
#include "ents.h"
#include "e_rick.h"
#include "e_sbonus.h"
#include "e_them.h"
#include "sound.h"
#include "screens.h"
#include "rects.h"
#include "kbd.h"
#include "parms.h"

U8 isave_frow;

/* FIXME this should be part of global.c */
U8 g_period = 0;
U8 g_waitkbd = G_FALSE;
rect_t *g_rects = NULL;


/*
 * local typedefs
 */
typedef enum {
  IMAIN, IMAP,
  PAUSE_PRESSED1, PAUSED, PAUSE_PRESSED2,
  PLAY0, PLAY1, PLAY2, PLAY3,
  CHAIN_SUBMAP, CHAIN_MAP, CHAIN_END,
  SCROLL_UP, SCROLL_DOWN,
  RESTART, GAMEOVER, GETNAME, EXIT
} cntl_t;


/*
 * local vars
 */
static cntl_t g_cntl;
static U8 speriod;

/*
 * prototypes
 */
static U8 scroll_up(void);
static U8 scroll_down(void);
static void frame(void);
static void init0(void);
static void init1(void);
static void play0(void);
static void play3(void);
static void restart(void);
static void isave(void);
static void irestore(void);


/*
 * Scroll up
 *
 */
static U8
scroll_up(void)
{
  U8 i, j;
  static U8 n = 0;

  /* translate map */
  for (i = MAP_ROW_SCRTOP; i < MAP_ROW_HBBOT; i++)
    for (j = 0x00; j < 0x20; j++)
      map_map[i][j] = map_map[i + 1][j];
  
  /* translate entities */
  for (i = 0; ent_ents[i].n != 0xFF; i++) {
    if (ent_ents[i].n) {
      ent_ents[i].ysave -= 8;
      ent_ents[i].trig_y -= 8;
      ent_ents[i].y -= 8;
      if (ent_ents[i].y & 0x8000) {  /* map coord. from 0x0000 to 0x0140 */
	/*sys_printf("DEBUG/scroll: entity %#04X is gone\n", i);*/
	ent_ents[i].n = 0;
      }
    }
  }

  /* display */
  d_map();
  ent_draw();
  d_cnt();
  map_frow++;

  /* period */
  if (n == 0) {
    speriod = g_period;
    g_period = 0;
  }

  /* loop */
  if (n++ == 7) {
    n = 0;
    /* activate visible entities */
    ent_actvis(map_frow + MAP_ROW_HBTOP, map_frow + MAP_ROW_HBBOT);
    
    /* prepare map */
    map_expand();
    
    /* display */
    d_map();
    ent_draw();
    d_cnt();

    /* period */
    g_period = speriod;
  }

  g_rects = &v_fbrect;
  return n;
}

/*
 * Scroll down
 *
 */
static U8
scroll_down(void)
{
  U8 i, j;
  static U8 n = 0;

  /* translate map */
  for (i = MAP_ROW_SCRBOT; i > MAP_ROW_HTTOP; i--)
    for (j = 0x00; j < 0x20; j++)
      map_map[i][j] = map_map[i - 1][j];
  
  /* translate entities */
  for (i = 0; ent_ents[i].n != 0xFF; i++) {
    if (ent_ents[i].n) {
      ent_ents[i].ysave += 8;
      ent_ents[i].trig_y += 8;
      ent_ents[i].y += 8;
      if (ent_ents[i].y > 0x0140) {  /* map coord. from 0x0000 to 0x0140 */
	/*sys_printf("DEBUG/scroll: entity %#04X is gone\n", i);*/
	ent_ents[i].n = 0;
      }
    }
  }
  
  /* display */
  d_map();
  ent_draw();
  d_cnt();
  map_frow--;

  /* period */
  if (n == 0) {
    speriod = g_period;
    g_period = 0;
  }

  /* loop */
  if (n++ == 7) {
    n = 0;
    /* activate visible entities */
    ent_actvis(map_frow + MAP_ROW_HTTOP, map_frow + MAP_ROW_HTBOT);

    /* prepare map */
    map_expand();

    /* display */
    d_map();
    ent_draw();
    d_cnt();

    /* period */
    g_period = speriod;
  }

  g_rects = &v_fbrect;
  return n;
}


/*
 * Main loop
 *
 */
void
mainloop(void)
{
  U32 tm, tmx;

  g_period = p_period ? p_period : G_PERIOD;

  tm = sys_gettime();
  g_cntl = IMAIN;

  /* perf. loop */
#ifdef DEBUG
  {
    int i;
    tmx = sys_gettime();
    for (i = 0; i < 100; i++)
      v_update(&v_fbrect);
    tm = sys_gettime();
    sys_printf("xrick/mainloop: period-min = %d\n", (tm-tmx)/100);
  }
#endif

  /* main loop */
  while (g_cntl != EXIT) {

    /*sys_printf("DEBUG/mainloop: ---- loop ----\n");*/

    /* timer */
    tmx = tm; tm = sys_gettime(); tmx = tm - tmx;
    if (tmx < g_period) sys_sleep(g_period - tmx);

    /* video */
    if (g_rects)
      v_update(g_rects);

    /* sound */

    /* events */
    if (g_waitkbd)
      v_eventsk();  /* process events, wait for a keyboard event */
    else
      v_events();  /* process events, return */

    /* frame */
    frame();

  }

}

/*
 * Prepare frame
 *
 */
static void
frame(void)
{
  while (1) {

    switch (g_cntl) {
    
    case IMAIN:                                 /* MAIN INTRO */
      if (scr_imain()) {
	init0();
	g_cntl = IMAP;
      }
      return;

    case IMAP:                                  /* MAP INTRO */
      if (scr_imap()) {
	init1();
	g_waitkbd = G_FALSE;
	g_cntl = PLAY0;
      }
      return;

    case PAUSE_PRESSED1:                        /* PAUSE PRESSED */
      if (!(k_status & K_PAUSE))
	g_cntl = PAUSED;
      return;

    case PAUSED:                                /* GAME PAUSED */
      if (k_status & K_PAUSE)
	g_cntl = PAUSE_PRESSED2;
      return;

    case PAUSE_PRESSED2:                        /* PAUSED PRESSED */
      if (!(k_status & K_PAUSE)) {
	g_cntl = PLAY2;
	g_waitkbd = G_FALSE;
      }
      return;

    case PLAY0:
      play0();
      break;
    
    case PLAY1:
      if (k_status & K_PAUSE) {
	snd_off();
	g_waitkbd = G_TRUE;
	g_cntl = PAUSE_PRESSED1;
      }
      else
	g_cntl = PLAY2;
      break;
    
    case PLAY2:
      if E_RICK_STTST(E_RICK_STDEAD) {
        if (--g_lives || g_cheat1)
	  g_cntl = RESTART;
        else
	  g_cntl = GAMEOVER; /* 0F */
      }
      else if (g_chsm)
	g_cntl = CHAIN_SUBMAP;
      else
	g_cntl = PLAY3;
      break;
    
    case PLAY3:
      play3();
      return;

    case CHAIN_SUBMAP:                          /* CHAIN SUBMAP */
      if (map_chain())
	g_cntl = CHAIN_END;
      else {
	g_bullets = 0x06;
	g_bombs = 0x06;
	g_map++;
	
	if (g_map == 0x04) {
	  /* FIXME @292?*/
	}
	
	g_cntl = CHAIN_MAP;
      }
      break;
    
    case CHAIN_MAP:                             /* CHAIN MAP */
      if (scr_imap()) {
	if (g_map >= 0x04)
	  g_cntl = GAMEOVER;
	else {
	  ent_ents[1].x = map_maps[g_map].x;
	  ent_ents[1].y = map_maps[g_map].y;
	  map_frow = map_maps[g_map].row;
	  g_submap = map_maps[g_map].submap;
	  g_cntl = CHAIN_END;
	}
	break;
      }
      else
	return;

    case CHAIN_END:                             /* CHAIN */
      map_init();
      isave();
      ent_clprev();
      d_map();
      d_cnt();
      g_rects = &v_fbrect;
      g_cntl = PLAY3;
      return;

    case SCROLL_UP:                             /* SCROLLUP */
      if (!scroll_up())
	g_cntl = PLAY0;
      return;
    
    case SCROLL_DOWN:                           /* SCROLL_DOWN */
      if (!scroll_down())
	g_cntl = PLAY0;
      return;

    case RESTART:                               /* RESTART AFTER DEATH */
      restart();
      g_cntl = PLAY0;
      return;

    case GAMEOVER:                              /* GAMEOVER SCREEN */
      if (scr_gameover())
	g_cntl = GETNAME;
      return;

    case GETNAME:                               /* GET NAME - HALL OF FAME */
      if (scr_getname()) 
	g_cntl = IMAIN;
      return;

    case EXIT:
      return;

    }
  }
}


/*
 * init0
 *
 */
static void
init0(void)
{
  E_RICK_STRST(0xff);

  g_lives = 6;
  g_bombs = 6;
  g_bullets = 6;
  g_score = 0;

  g_map = p_map;
    
  ent_ents[1].x = map_maps[g_map].x;
  ent_ents[1].y = map_maps[g_map].y;
  ent_ents[1].w = 0x18;
  ent_ents[1].h = 0x15;
  ent_ents[1].n = 0x01;
  ent_ents[1].sprite = 0x01;
  ent_ents[ENT_ENTSNUM].n = 0xff;

  map_resetmarks();
}


/*
 * init1
 *
 */
static void
init1(void)
{
  v_clear();

  if (p_submap != 0) {
    g_submap = p_submap;
    map_frow = p_frow;
  }
  else {
    g_submap = map_maps[g_map].submap;
    map_frow = map_maps[g_map].row;
  }

  map_init();
  isave();
  d_map();
  d_cnt();
  g_rects = &v_fbrect;
}

/*
 * play0
 *
 */
static void
play0(void)
{
  /* clear */
  g_rects = NULL;

  /* end */
  if (k_status & K_END) {
    g_cntl = GAMEOVER; /* FF */
    return;
  }

  /* exit */
  if (k_last == K_XTRA) {
    g_cntl = EXIT;
    return;
  }

  /* action */
  ent_action();
  /* call 0270 */
  e_them_rndseed++;

  /* speed bonus counter (01A6) */
  if (e_sbonus_counting && --e_sbonus_counter == 0) {
    e_sbonus_counter = 0x1e;
    if (e_sbonus_bonus)
      e_sbonus_bonus--;
  }

  g_cntl = PLAY1;
}


/*
 * play3
 *
 */
static void
play3(void)
{
  /* FIXME should use constants here */
  static rect_t rcnt = { 0x28, 0x08, 0xf0, 8, NULL };
  static rect_t *r;

  d_cnt_clr();

  ent_draw();

  /*
    snd_something();
    v_sync;
    snd_something();
    v_sync;
  */

  /*
   * original rick just calls d_cnt_updbb but they have another way of
   * implementing clipping (via a call to 0d2f in 0a54). We have to do
   * the clipping by ourselves -> keep redrawing everything.
   */
  d_cnt();
  r = &rcnt; r->next = ent_rects;
  g_rects = r;

  if (!E_RICK_STTST(E_RICK_STZOMBIE)) {
    if (ent_ents[1].y >= 0xcc) {
      g_cntl = SCROLL_UP;
      return;
    }
    if (ent_ents[1].y <= 0x60) {
      g_cntl = SCROLL_DOWN;
      return;
    }
  }  

  g_cntl = PLAY0;
}


/*
 * restart
 *
 */
static void
restart(void)
{
  /*sys_printf("DEBUG/xrick: restart\n");*/

  E_RICK_STRST(E_RICK_STDEAD|E_RICK_STZOMBIE);

  g_bullets = 6;
  g_bombs = 6;

  ent_ents[1].n = 0x01;

  irestore();
  map_init();
  isave();
  ent_clprev();
  d_map();
  d_cnt();
  g_rects = &v_fbrect;
}


/*
 * isave (0bbb)
 *
 */
static void
isave(void)
{
  e_rick_save();
  isave_frow = map_frow;
}


/*
 * irestore (0bdc)
 *
 */
static void
irestore(void)
{
  e_rick_restore();
  map_frow = isave_frow;
}


/* eof */
