/*
 * xrick/src/scr_imap.c
 *
 * This source code file is part of the xrick distribution, and as such is
 * copyrighted. Please read the README file in the root directory of the
 * distribution for information about copyrights, trademarks, license,
 * disclaimers, etc.
 */

#include <stdio.h>

#include "system.h"
#include "global.h"
#include "screens.h"

#include "global.h"
#include "rects.h"
#include "video.h"
#include "draw.h"
#include "kbd.h"

/*
 * local vars
 */
static U16 step;              /* current step */
static U16 count;             /* number of loops for current step */
static U16 run;               /* 1 = run, 0 = no more step */
static U8 flipflop;           /* flipflop for top, bottom, left, right */
static U16 spnum;             /* sprite number */
static U16 spx, spdx;         /* sprite x position and delta */
static U16 spy, spdy;         /* sprite y position and delta */
static U16 spbase, spoffs;    /* base, offset for sprite numbers table */
static U8 seq = 0;            /* anim sequence */

/*static rect_t anim_rect = { 120, 16, 184, 80, NULL };*/ /* anim rectangle */
static rect_t anim_rect = { 120, 16, 64, 64, NULL }; /* anim rectangle */

/*
 * prototypes
 */
static void drawtb(void);
static void drawlr(void);
static void drawsprite(void);
static void drawcenter(void);
static void nextstep(void);
static void anim(void);
static void init(void);

/*
 * Map introduction
 *
 * AS: 1948
 *
 * return: 0 when running, 1 when finished.
 */
U8
scr_imap(void)
{
  switch (seq) {
  case 0:  /* display text etc. */
    /*sys_printf("xrick/screens: start map intro for map %d\n", g_map);*/
    
    v_clear();

    d_tlbank = 1;
    d_tllst = scr_imaptext[g_map];
    d_setfbp(32, 0);
    d_filter = 0xaaaa;
    d_tl_sblst();

    d_setfbp(32, 96);
    d_filter = 0x5555;
    d_tl_lst();

    g_rects = &v_fbrect;

    d_filter = 0xffff;

    init();
    nextstep();
    drawcenter();
    drawtb();
    drawlr();
    drawsprite();
    k_last = 0;

    seq = 1;
    break;
  case 1:  /* top and bottom borders */
    drawtb();
    g_rects = &anim_rect;
    seq = 2;
    break;
  case 2:  /* background and sprite */
    anim();
    drawcenter();
    drawsprite();
    g_rects = &anim_rect;
    seq = 3;
    break;
  case 3:  /* all borders */
    drawtb();
    drawlr();
    g_rects = &anim_rect;
    seq = 1;
    break;
  case 4:  /* wait for key release */
    g_rects = NULL;
    if (!(k_status & K_FIRE))
      seq = 5;
    else
      sys_sleep(50); /* .5s */
    break;
  }

  if (k_status & K_FIRE) {  /* end as soon as key pressed */
    seq = 4;
  }

  if (seq == 5) {  /* end as soon as key pressed */
    v_clear();
    g_rects = &v_fbrect;
    seq = 0;
    /*sys_printf("xrick/screens: end intro for map\n");*/
    return 1;
  }
  else
    return 0;
}


/*
 * Display top and bottom borders (0x1B1F)
 *
 */
static void
drawtb(void)
{
  U8 i;

  flipflop++;
  if (flipflop & 0x01) {
    d_setfbp(128, 16);
    for (i = 0; i < 6; i++)
      d_tile(0x40);
    d_setfbp(128, 72);
    for (i = 0; i < 6; i++)
      d_tile(0x06);
  }
  else {
    d_setfbp(128, 16);
    for (i = 0; i < 6; i++)
      d_tile(0x05);
    d_setfbp(128, 72);
    for (i = 0; i < 6; i++)
      d_tile(0x40);
  }
}


/*
 * Display left and right borders (0x1B7C)
 *
 */
static void
drawlr(void)
{
  U8 i;

  if (flipflop & 0x02) {
    for (i = 0; i < 8; i++) {
      d_setfbp(120, 16 + i * 8);
      d_tile(0x04);
      d_setfbp(176, 16 + i * 8);
      d_tile(0x04);
    }
  }
  else {
    for (i = 0; i < 8; i++) {
      d_setfbp(120, 16 + i * 8);
      d_tile(0x2b);
      d_setfbp(176, 16 + i * 8);
      d_tile(0x2b);
    }
  }
}


/*
 * Draw the sprite (0x19C6)
 *
 */
static void
drawsprite(void)
{
#ifdef GFXPC
  d_sprite(2 * (spx & 0x01), spnum,
	      128 + ((spx << 1) & 0x1c), 24 + (spy << 1));
#endif
#ifdef GFXST
  d_sprite(spnum, 128 + ((spx << 1) & 0x1c), 24 + (spy << 1));
#endif
}


/*
 * Draw the background (0x1AF1)
 *
 */
static void
drawcenter(void)
{
  static U8 tn0[] = { 0x07, 0x5b, 0x7f, 0xa3, 0xc7 };
  U8 i, j, tn;

  tn = tn0[g_map];
  for (i = 0; i < 6; i++) {
    d_setfbp(128, (24 + 8 * i));
    for (j = 0; j < 6; j++)
      d_tile(tn++);
  }
}


/*
 * Next Step (0x1A74)
 *
 */
static void
nextstep(void)
{
  if (scr_imapsteps[step].count) {
    count = scr_imapsteps[step].count;
    spdx = scr_imapsteps[step].dx;
    spdy = scr_imapsteps[step].dy;
    spbase = scr_imapsteps[step].base;
    spoffs = 0;
    step++;
  }
  else {
    run = 0;
  }
}


/*
 * Anim (0x1AA8)
 *
 */
static void
anim(void)
{
  U16 i;

  if (run) {
    i = scr_imapsl[spbase + spoffs];
    if (i == 0) {
      spoffs = 0;
      i = scr_imapsl[spbase];
    }
    spnum = i;
    spoffs++;
    spx += spdx;
    spy += spdy;
    count--;
    if (count == 0)
      nextstep();
  }
}


/*
 * Initialize (0x1A43)
 *
 */
static void
init(void)
{
  run = 0; run--;
  step = scr_imapsofs[g_map];
  spx = scr_imapsteps[step].dx;
  spy = scr_imapsteps[step].dy;
  step++;
  spnum = 0; /* NOTE spnum in [8728] is never initialized ? */
}

/* eof */



